#!/usr/bin/env python2

import argparse
from ctypes import cdll
import multiprocessing
from multiprocessing.pool import ThreadPool
import os
import re
import select
import socket
import struct
import subprocess
import sys
import time

import requests


parser = argparse.ArgumentParser()
parser.add_argument("--host", default="mon")
parser.add_argument("--port", default="80")
parser.add_argument("--arp_count", default=3, type=int)
parser.add_argument("--timeout", default=1.0, type=float)
parser.add_argument("--processes", default=multiprocessing.cpu_count(), type=int)
args = parser.parse_args()

PROMETHEUS_QUERY_API = "http://{}:{}/api/v1/query".format(args.host, args.port)

libc = cdll.LoadLibrary('libc.so.6')
setns = libc.setns

HOSTNAME = socket.gethostname()

# socket timeout
TIMEOUT = args.timeout

# Ethernet
TYPEFRAME = 0x0806

# ARP packet
TYPEHRD = 1

# protocol ip
PROTOCOLTYPE = 0x0800

# ARP default
PACKETSIZE = 42

ARPREQUEST = 1
ARPREPLY = 2


class ARPSendPacket(object):

    @staticmethod
    def _convert_mac(mac):
        macbin = ''
        for l in re.split(r':', mac):
            macbin += chr(int('0x' + l, 16))
        return macbin

    def __init__(self, eth_dest, eth_src, ip_sedr, mac_sedr, ip_recvr, mac_recvr):
        # ethernet header
        self._eth_dest = self._convert_mac(eth_dest)
        self._eth_src = self._convert_mac(eth_src)
        self._type_frame = struct.pack('H', socket.htons(TYPEFRAME))

        # ARP packet
        self._type_hrd = struct.pack('H', socket.htons(TYPEHRD))
        self._type_pro = struct.pack('H', socket.htons(PROTOCOLTYPE))
        self._mac_len = struct.pack('B', struct.calcsize('6B'))
        self._op = struct.pack('H', socket.htons(ARPREQUEST))
        self._mac_sedr = self._convert_mac(mac_sedr)
        self._ip_sedr = socket.inet_aton(ip_sedr)
        self._mac_recvr = self._convert_mac(mac_recvr)
        self._ip_recvr = socket.inet_aton(ip_recvr)
        self._ip_len = struct.pack('B', len(self._ip_sedr))

    def __str__(self):
        return self._eth_dest + self._eth_src + self._type_frame + \
               self._type_hrd + self._type_pro + self._mac_len + \
               self._ip_len + self._op + self._mac_sedr + \
               self._ip_sedr + self._mac_recvr + self._ip_recvr


class Namespace(object):

    @staticmethod
    def _get_ns_path(nspath=None, nsname=None, nspid=None):
        if nsname:
            nspath = '/run/netns/%s' % nsname
        elif nspid:
            nspath = '/proc/%d/ns/net' % nspid
        return nspath

    def __init__(self, nsname=None, nspath=None, nspid=None):
        self.mypath = self._get_ns_path(nspid=os.getpid())
        self.targetpath = self._get_ns_path(nspath,
                                            nsname=nsname,
                                            nspid=nspid)
        if not self.targetpath:
            raise ValueError('invalid namespace')

    def __enter__(self):
        self.myns = open(self.mypath)
        with open(self.targetpath) as fd:
            setns(fd.fileno(), 0)

    def __exit__(self, *args):
        setns(self.myns.fileno(), 0)
        self.myns.close()


def call_process(cmd):
    p = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output = p.communicate()
    return output


def recv_arp_packets(soc, ip_bin, ip_recvrs, timeout):
    time_left = timeout
    packets = set()
    ips_left = ip_recvrs.copy()
    while True:
        time_start = time.time()
        srecv = select.select([soc], [], [], time_left)
        time_received = time.time()
        time_left -= time_received - time_start
        if not srecv[0]:
            return packets
        try:
            data = soc.recv(PACKETSIZE)
        except OSError:
            return packets
        if len(data) == PACKETSIZE and ord(data[21]) == ARPREPLY:
            ip_to = data[38:42]
            if ip_to == ip_bin:
                ip_from = socket.inet_ntoa(data[28:32])
                if ip_from in ips_left:
                    packets.add(data)
                    ips_left.discard(ip_from)
        if time_left <= 0 or len(ips_left) == 0:
            return packets


def make_subnet_arping(namespace, interface, mac_sedr, ip_sedr, ip_recvrs):
    with Namespace(nsname=namespace):
        soc = socket.socket(socket.PF_PACKET, socket.SOCK_RAW)
        try:
            soc.bind((interface, TYPEFRAME))
        except OSError:
            return set(), ip_recvrs
        for ip_recvr in ip_recvrs:
            bc_packet = ARPSendPacket('ff:ff:ff:ff:ff:ff', mac_sedr, ip_sedr, mac_sedr, ip_recvr, '00:00:00:00:00:00')
            try:
                soc.send(bc_packet.__str__())
            except OSError:
                return set(), ip_recvrs
        packets = recv_arp_packets(soc, socket.inet_aton(ip_sedr), ip_recvrs, TIMEOUT)
        res_addrs = set()
        for packet in packets:
            ip_from = packet[28:32]
            mac_from = packet[6:12]
            res_addrs.add((ip_from, mac_from))
        for _ in range(args.arp_count - 1):
            tmp_res_addrs = set()
            str_ips = set()
            for addr in res_addrs:
                ip_str = socket.inet_ntoa(addr[0])
                str_ips.add(ip_str)
                mac_str = ':'.join(x.encode('hex') for x in addr[1])
                uc_packet = ARPSendPacket(mac_str, mac_sedr, ip_sedr, mac_sedr, ip_str, mac_str)
                try:
                    soc.send(uc_packet.__str__())
                except OSError:
                    return set(), ip_recvrs
            packets = recv_arp_packets(soc, socket.inet_aton(ip_sedr), str_ips, TIMEOUT)
            for packet in packets:
                ip_from = packet[28:32]
                mac_from = packet[6:12]
                tmp_res_addrs.add((ip_from, mac_from))
            res_addrs = tmp_res_addrs
        ips = set()
        for addr in res_addrs:
            ips.add(socket.inet_ntoa(addr[0]))
        soc.close()
        return ips, ip_recvrs.difference(ips)


def instant_query(expression):
    params = {"query": expression}
    result = requests.get(PROMETHEUS_QUERY_API, params=params).json()["data"]["result"]
    return result


def parse_ip_a(data):

    # Expected data format:
    # 1: lo: <LOOPBACK,UP,LOWER_UP> mtu 65536 qdisc noqueue state UNKNOWN group default qlen 1000
    #     link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
    #     inet 127.0.0.1/8 scope host lo
    #        valid_lft forever preferred_lft forever
    #     inet6 ::1/128 scope host
    #        valid_lft forever preferred_lft forever

    interfaces = {}
    name = ""
    mac = ""
    addresses = []
    for line in data.split('\n'):
        m = re.match(r"(\d+):\s?(.+):", line)
        if m:
            interfaces[name] = {"addresses": addresses, "mac": mac}
            mac = ""
            addresses = []
            name = m.group(2)
        m = re.match(r"\s+link/\w+\s([0-9a-f:]+)", line)
        if m:
            mac = m.group(1)
        m = re.match(r"\s+inet\s([0-9.]+)/(\d+)", line)
        if m:
            addresses.append((m.group(1), m.group(2)))
    interfaces[name] = {"addresses": addresses, "mac": mac}
    interfaces.pop("", None)
    interfaces.pop("lo", None)
    return interfaces


def get_ip4_network(ip, mask):
    host_bits = 32 - int(mask)
    netmask = socket.inet_ntoa(struct.pack('!I', (1 << 32) - (1 << host_bits)))
    subnet = ".".join(map(str, [i & m for i, m in zip(map(int, ip.split(".")), map(int, netmask.split(".")))]))
    return subnet


def gather():
    hosted_networks = set()
    dhcp_agents = instant_query("openstack_neutron_agent_dhcp_metadata")
    for agent in dhcp_agents:
        host = agent["metric"]["hostname"]
        if host == HOSTNAME:
            network_id = agent["metric"]["network_id"]
            hosted_networks.add(network_id)

    target_networks = set()
    dhcp_subnets = set()
    dhcp_enabled = instant_query("openstack_neutron_subnet_enable_dhcp")
    for net in dhcp_enabled:
        enabled = net["value"][1]
        if enabled == "1":
            subnet_id = net["metric"]["id"]
            dhcp_subnets.add(subnet_id)
            network_id = net["metric"]["network_id"]
            if network_id in hosted_networks:
                target_networks.add(network_id)

    active_instances = set()
    instances = instant_query("openstack_nova_instance_status")
    for instance in instances:
        status = instance["value"][1]
        if status == "0":
            instance_id = instance["metric"]["id"]
            active_instances.add(instance_id)

    thread_pool = ThreadPool(multiprocessing.cpu_count())
    ns_checks = []
    for network_id in target_networks:
        cmd = "ip netns exec qdhcp-{} ip address".format(network_id)
        result = thread_pool.apply_async(call_process, (cmd,))
        ns_checks.append({"result": result, "network_id": network_id})
    thread_pool.close()
    thread_pool.join()

    namespaces_ifs = {}
    for ns_check in ns_checks:
        network_id = ns_check["network_id"]
        out, err = ns_check["result"].get()
        namespaces_ifs[network_id] = parse_ip_a(out)

    ports = instant_query("openstack_neutron_port_metadata")
    check_map = {}
    mapping = {}
    for port in ports:
        network_id = port["metric"].get("network_id", "")
        if network_id not in hosted_networks:
            continue
        subnet_id = port["metric"].get("subnet_id", "")
        if subnet_id not in dhcp_subnets:
            continue
        device_id = port["metric"].get("device_id", "")
        if device_id not in active_instances:
            continue
        ip_address = port["metric"].get("ip_address", "")
        if len(ip_address.split('.')) != 4:
            continue

        ns_ifs = namespaces_ifs[network_id]
        if len(ns_ifs) == 0:
            continue
        name = ns_ifs.keys()[0]
        interface = ns_ifs[name]
        mac = interface.get("mac", "")
        addresses = interface.get("addresses", [])
        ns = "qdhcp-{}".format(network_id)
        for addr in addresses:
            a_ip = addr[0]
            mask = addr[1]
            net = get_ip4_network(a_ip, mask)
            target_net = get_ip4_network(ip_address, mask)
            if net == target_net:
                id = ns + " " + name + " " + mac + " " + a_ip
                if id not in check_map:
                    check_map[id] = set()
                check_map[id].add(ip_address)
                id_long = id + " " + ip_address
                mapping[id_long] = {"network_id": network_id, "device_id": device_id, "ip": ip_address}
                break
    checks = []
    pool = ThreadPool(args.processes)
    for id in check_map:
        ns, name, mac, a_ip = id.split(' ')
        result = pool.apply_async(make_subnet_arping, (ns, name, mac, a_ip, check_map[id]))
        checks.append({"result": result, "id": id})
    pool.close()
    pool.join()
    for check in checks:
        id = check["id"]
        try:
            passed, failed = check["result"].get()
        except IOError:
            passed, failed = set(), check_map[id]
        for ip in passed:
            id_long = id + " " + ip
            res = mapping[id_long]
            print "instance_arping,network_id=%(network_id)s,ip_address=%(ip_address)s,id=%(id)s success=%(success)s" % \
                  {
                      'network_id': res['network_id'],
                      'ip_address': res['ip'],
                      'id': res['device_id'],
                      'success': "1"
                  }
        for ip in failed:
            id_long = id + " " + ip
            res = mapping[id_long]
            print "instance_arping,network_id=%(network_id)s,ip_address=%(ip_address)s,id=%(id)s success=%(success)s" % \
                  {
                      'network_id': res['network_id'],
                      'ip_address': res['ip'],
                      'id': res['device_id'],
                      'success': "0"
                  }
if __name__ == "__main__":
    try:
        gather()
        print "instance_arping check_up=1"
    except Exception as e:
        print "instance_arping check_up=0"
        sys.stderr.write(str(e))
        sys.exit(1)
